import { Test, TestingModule } from '@nestjs/testing';
import { ReposService } from '../repos/repos.service';
import { PipelineUnits } from './enums/pipeline-units.enum';
import { PipelineTask } from './pipeline-task.entity';
import { Pipeline } from '../pipelines/pipeline.entity';
import { Project } from '../projects/project.entity';
import { TaskStatuses } from './enums/task-statuses.enum';
import { getLoggerToken, PinoLogger } from 'nestjs-pino';
import { PipelineTaskRunner } from './pipeline-task.runner';
import { WorkUnitMetadata } from './models/work-unit-metadata.model';
import { AmqpConnection } from '@golevelup/nestjs-rabbitmq';
describe('PipelineTaskRunner', () => {
  let runner: PipelineTaskRunner;
  let reposService: ReposService;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: ReposService,
          useValue: {
            getWorkspaceRootByTask: () => 'workspace-root',
            checkout: async () => undefined,
          },
        },
        {
          provide: getLoggerToken(PipelineTaskRunner.name),
          useValue: new PinoLogger({}),
        },
        {
          provide: 'spawn',
          useValue: () => undefined,
        },
        PipelineTaskRunner,
        {
          provide: AmqpConnection,
          useValue: {},
        },
      ],
    }).compile();

    reposService = module.get(ReposService);
    runner = module.get(PipelineTaskRunner);
  });

  it('should be defined', () => {
    expect(runner).toBeDefined();
  });

  it('onNewTask', async () => {
    const task = new PipelineTask();
    let tmpTask;
    const doTask = jest
      .spyOn(runner, 'doTask')
      .mockImplementation(async (task) => {
        tmpTask = task;
      });
    await runner.onNewTask(task);
    expect(tmpTask).toEqual(task);
    expect(doTask).toBeCalledTimes(1);
  });

  describe('test biz', () => {
    let emitEvent: jest.SpyInstance;
    beforeEach(() => {
      emitEvent = jest
        .spyOn(runner, 'emitEvent')
        .mockImplementation((..._) => Promise.resolve());
    });

    describe('doTask', () => {
      let checkout: jest.SpyInstance;
      let doTaskUnit: jest.SpyInstance;

      beforeEach(() => {
        checkout = jest
          .spyOn(runner, 'checkout')
          .mockImplementation((..._) => Promise.resolve('/null'));
        doTaskUnit = jest
          .spyOn(runner, 'doTaskUnit')
          .mockImplementation((..._) => Promise.resolve());
      });

      it('only checkout', async () => {
        const task = new PipelineTask();
        (task.id = 'taskId'), (task.pipeline = new Pipeline());
        task.units = [PipelineUnits.checkout];
        task.pipeline.id = 'pipelineId';
        task.pipeline.project = new Project();
        task.pipeline.project.id = 'projectId';
        task.pipeline.workUnitMetadata = new WorkUnitMetadata();
        task.pipeline.workUnitMetadata.version = 1;
        task.pipeline.workUnitMetadata.units = [
          {
            type: PipelineUnits.checkout,
            scripts: [],
          },
        ];

        await runner.doTask(task);

        expect(checkout).toBeCalledTimes(1);
        expect(doTaskUnit).toBeCalledTimes(0);
        expect(emitEvent).toBeCalledTimes(2);
        expect(emitEvent.mock.calls[0][0]).toMatchObject(task);
        expect(emitEvent.mock.calls[0][1]).toBeNull();
        expect(emitEvent.mock.calls[0][2]).toEqual(TaskStatuses.working);
        expect(emitEvent.mock.calls[1][0]).toMatchObject(task);
        expect(emitEvent.mock.calls[1][1]).toBeNull();
        expect(emitEvent.mock.calls[1][2]).toEqual(TaskStatuses.success);
      });

      it('many units', async () => {
        const task = new PipelineTask();
        (task.id = 'taskId'), (task.pipeline = new Pipeline());
        task.units = [
          PipelineUnits.checkout,
          PipelineUnits.test,
          PipelineUnits.deploy,
        ];
        task.pipeline.id = 'pipelineId';
        task.pipeline.project = new Project();
        task.pipeline.project.id = 'projectId';
        task.pipeline.workUnitMetadata = new WorkUnitMetadata();
        task.pipeline.workUnitMetadata.version = 1;
        task.pipeline.workUnitMetadata.units = [
          {
            type: PipelineUnits.checkout,
            scripts: [],
          },
          {
            type: PipelineUnits.installDependencies,
            scripts: ['pwd'],
          },
          {
            type: PipelineUnits.test,
            scripts: ['pwd'],
          },
          {
            type: PipelineUnits.deploy,
            scripts: ['pwd', 'uname'],
          },
        ];

        await runner.doTask(task);

        expect(checkout).toBeCalledTimes(1);
        expect(doTaskUnit).toBeCalledTimes(1);
        expect(emitEvent).toBeCalledTimes(2);
      });

      it('unit work failed', async () => {
        const task = new PipelineTask();
        (task.id = 'taskId'), (task.pipeline = new Pipeline());
        task.units = [PipelineUnits.checkout, PipelineUnits.test];
        task.pipeline.id = 'pipelineId';
        task.pipeline.project = new Project();
        task.pipeline.project.id = 'projectId';
        task.pipeline.workUnitMetadata = new WorkUnitMetadata();
        task.pipeline.workUnitMetadata.version = 1;
        task.pipeline.workUnitMetadata.units = [
          {
            type: PipelineUnits.checkout,
            scripts: [],
          },
          {
            type: PipelineUnits.test,
            scripts: ['pwd'],
          },
        ];

        doTaskUnit = jest
          .spyOn(runner, 'doTaskUnit')
          .mockImplementation((..._) =>
            Promise.reject(new Error('test error')),
          );
        await runner.doTask(task);

        expect(checkout).toBeCalledTimes(1);
        expect(doTaskUnit).toBeCalledTimes(1);
        expect(emitEvent).toBeCalledTimes(2);
        expect(emitEvent.mock.calls[1][0]).toMatchObject(task);
        expect(emitEvent.mock.calls[1][1]).toBeNull();
        expect(emitEvent.mock.calls[1][2]).toEqual(TaskStatuses.failed);
      });
    });

    describe('doTaskUnit', () => {
      it('success', async () => {
        const runScript = jest
          .spyOn(runner, 'runScript')
          .mockImplementation((..._) => Promise.resolve());
        const task = new PipelineTask();

        const unit = PipelineUnits.test;
        const workspacePath = '/null';
        await runner.doTaskUnit(unit, ['pwd'], task, workspacePath);

        expect(emitEvent.mock.calls[0][0]).toEqual(task);
        expect(emitEvent.mock.calls[0][1]).toEqual(unit);
        expect(emitEvent.mock.calls[0][2]).toEqual(TaskStatuses.working);
        expect(emitEvent.mock.calls[1][0]).toEqual(task);
        expect(emitEvent.mock.calls[1][1]).toEqual(unit);
        expect(emitEvent.mock.calls[1][2]).toEqual(TaskStatuses.success);
        expect(runScript.mock.calls[0][0]).toEqual('pwd');
        expect(runScript.mock.calls[0][1]).toEqual(workspacePath);
        expect(runScript.mock.calls[0][2]).toEqual(task);
        expect(runScript.mock.calls[0][3]).toEqual(unit);
      });
      it('failed', async () => {
        const runScript = jest
          .spyOn(runner, 'runScript')
          .mockImplementation((..._) =>
            Promise.reject(new Error('test error')),
          );
        const task = new PipelineTask();

        const unit = PipelineUnits.test;
        const workspacePath = '/null';
        await expect(
          runner.doTaskUnit(unit, ['pwd'], task, workspacePath),
        ).rejects.toThrow('test error');

        expect(emitEvent.mock.calls[1]?.[0]).toEqual(task);
        expect(emitEvent.mock.calls[1]?.[1]).toEqual(unit);
        expect(emitEvent.mock.calls[1]?.[2]).toEqual(TaskStatuses.failed);
        expect(runScript).toBeCalledTimes(1);
      });
    });

    describe('runScript', () => {
      it('normal', async () => {
        const spawn = jest.fn((..._: any[]) => ({
          stdout: {
            on: () => undefined,
          },
          stderr: {
            on: () => undefined,
          },
          addListener: (_: any, fn: (code: number) => void) => {
            fn(0);
          },
        }));
        (runner as any).spawn = spawn;

        const task = new PipelineTask();
        task.id = 'taskId';
        const unit = PipelineUnits.deploy;

        await runner.runScript('script name', 'workspaceRoot', task, unit);
        expect(spawn).toHaveBeenCalledTimes(1);
        expect(spawn.mock.calls[0][0]).toEqual('script name');
        expect(spawn.mock.calls[0][1]).toMatchObject({
          shell: true,
          cwd: 'workspaceRoot',
        });
      });
      it('failed', async () => {
        const spawn = jest.fn((..._: any[]) => ({
          stdout: {
            on: () => undefined,
          },
          stderr: {
            on: () => undefined,
          },
          addListener: (_: any, fn: (code: number) => void) => {
            fn(1);
          },
        }));
        (runner as any).spawn = spawn;

        const task = new PipelineTask();
        task.id = 'taskId';
        const unit = PipelineUnits.deploy;

        expect(
          runner.runScript('script name', 'workspaceRoot', task, unit),
        ).rejects.toThrowError();
      });
      it('wait emit message done', async () => {
        let finishedFn: () => void;
        const on = jest.fn((_: any, fn: (buff: Buffer) => void) => {
          setTimeout(() => {
            fn(Buffer.from('message 1'));
            setTimeout(() => {
              fn(Buffer.from('message 2'));
              setTimeout(() => {
                fn(Buffer.from('message 3'));
                finishedFn();
              }, 1000);
            }, 10);
          }, 10);
        });
        const spawn = jest.fn((..._: any[]) => ({
          stdout: {
            on,
          },
          stderr: {
            on,
          },
          addListener: (_: any, fn: (code: number) => void) => {
            finishedFn = () => fn(0);
          },
        }));

        let emitSuccessCount = 0;
        jest.spyOn(runner, 'emitEvent').mockImplementation((..._: any[]) => {
          return new Promise((resolve) => {
            setTimeout(() => {
              emitSuccessCount++;
              resolve();
            }, 1000);
          });
        });
        (runner as any).spawn = spawn;

        const task = new PipelineTask();
        task.id = 'taskId';
        const unit = PipelineUnits.deploy;

        await runner.runScript('script name', 'workspaceRoot', task, unit);
        expect(emitSuccessCount).toEqual(1 + 6);
      });
    });
  });
});
