import { ListLogsOption } from './models/list-logs.options';
import { PipelineTask } from './../pipeline-tasks/pipeline-task.entity';
import { Injectable, NotFoundException } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { F_OK } from 'constants';
import { access, mkdir } from 'fs/promises';
import { join } from 'path';
import { gitP } from 'simple-git';
import { Repository } from 'typeorm';
import { Project } from '../projects/project.entity';
import { ListBranchesArgs } from './dtos/list-branches.args';
import { ConfigService } from '@nestjs/config';
import { Commit } from './dtos/log-list.model';
import { Nack, RabbitRPC, RabbitSubscribe } from '@golevelup/nestjs-rabbitmq';
import { Pipeline } from '../pipelines/pipeline.entity';
import { InjectPinoLogger, PinoLogger } from 'nestjs-pino';
import {
  EXCHANGE_REPO,
  QUEUE_FETCH,
  QUEUE_LIST_COMMITS,
  QUEUE_REFRESH_REPO,
  ROUTE_FETCH,
  ROUTE_LIST_COMMITS,
} from './repos.constants';
import { getSelfInstanceQueueKey } from '../commons/utils/rabbit-mq';
import {
  getInstanceName,
  getSelfInstanceRouteKey,
} from '../commons/utils/rabbit-mq';
import { ApplicationException } from '../commons/exceptions/application.exception';
import {
  EXCHANGE_PROJECT_FANOUT,
  ROUTE_PROJECT_CHANGE,
} from '../projects/projects.constants';
import { RedisMutexService } from '../commons/redis-mutex/redis-mutex.service';
import { rm } from 'fs/promises';

const DEFAULT_REMOTE_NAME = 'origin';
const INFO_PATH = '@info';
@Injectable()
export class ReposService {
  constructor(
    @InjectRepository(Project)
    private readonly projectRepository: Repository<Project>,
    private readonly configService: ConfigService,
    @InjectPinoLogger(ReposService.name)
    private readonly logger: PinoLogger,
    private readonly redisMutexService: RedisMutexService,
  ) {}

  getWorkspaceRoot(project: Project): string {
    return join(
      this.configService.get<string>('workspaces.root'),
      encodeURIComponent(project.name),
      INFO_PATH,
    );
  }

  getDeployRoot(task: PipelineTask) {
    return join(
      this.configService.get<string>('workspaces.root'),
      encodeURIComponent(task.pipeline.project.name),
      encodeURIComponent(`deploy-${task.pipeline.name}`),
    );
  }

  async getGit(
    project: Project,
    workspaceRoot?: string,
    { fetch = true } = {},
  ) {
    if (!workspaceRoot) {
      workspaceRoot = this.getWorkspaceRoot(project);
    }

    await access(workspaceRoot, F_OK).catch(async () => {
      await mkdir(workspaceRoot, { recursive: true });
    });
    const git = gitP(workspaceRoot);
    if (!(await git.checkIsRepo().catch(() => false))) {
      await git.init();
      await git.addRemote(DEFAULT_REMOTE_NAME, project.sshUrl);
    }
    if (fetch) {
      await git.fetch();
    }
    return git;
  }

  async listLogs({ project, branch }: ListLogsOption) {
    const git = await this.getGit(project);
    return await git.log(
      branch ? ['--branches', `remotes/origin/${branch}`, '--'] : ['--all'],
    );
  }

  async listBranches(dto: ListBranchesArgs) {
    const project = await this.projectRepository.findOneOrFail({
      id: dto.projectId,
    });
    const git = await this.getGit(project);
    return git.branch();
  }

  async checkout(task: PipelineTask, workspaceRoot: string) {
    const git = await this.getGit(task.pipeline.project, workspaceRoot);
    try {
      await git.fetch(DEFAULT_REMOTE_NAME);
    } catch (err) {
      if (err.message.includes("couldn't find remote ref nonexistent")) {
        throw new NotFoundException(err.message);
      }
      throw err;
    }
    await git.checkout([task.commit]);
  }

  /**
   * get workspace root absolute path
   *
   * ! example: `/var/tmp/fennec-workspaces-root/project/pipeline_name-commit_hash`
   * @param task {PipelineTask} task (with pipeline and project info)
   */
  getWorkspaceRootByTask(task: PipelineTask) {
    return join(
      this.configService.get<string>('workspaces.root'),
      encodeURIComponent(task.pipeline.project.name),
      encodeURIComponent(`${task.pipeline.name}-${task.commit}`),
    );
  }

  async checkout4Task(task: PipelineTask): Promise<string> {
    const path = this.getWorkspaceRootByTask(task);
    await this.checkout(task, path);
    return path;
  }

  @RabbitRPC({
    exchange: EXCHANGE_REPO,
    routingKey: [
      ROUTE_LIST_COMMITS,
      getSelfInstanceRouteKey(ROUTE_LIST_COMMITS),
    ],
    queue: getSelfInstanceQueueKey(QUEUE_LIST_COMMITS),
    queueOptions: {
      autoDelete: true,
    },
  })
  async listCommits(pipeline: Pipeline): Promise<[Error, Commit[]?]> {
    const git = await this.getGit(pipeline.project, undefined, {
      fetch: false,
    });
    try {
      const data = await git.log([
        '-100',
        '--branches',
        `remotes/origin/${pipeline.branch}`,
        '--',
      ]);
      return [
        null,
        data.all.map(
          (it) =>
            ({
              ...it,
              date: new Date(it.date),
            } as Commit),
        ),
      ];
    } catch (error) {
      this.logger.error(
        { error, pipeline },
        '[listCommits] %s',
        error?.message,
      );
      return [new ApplicationException(error)];
    }
  }

  @RabbitRPC({
    exchange: EXCHANGE_REPO,
    routingKey: [ROUTE_FETCH, getSelfInstanceRouteKey(ROUTE_FETCH)],
    queue: getSelfInstanceQueueKey(QUEUE_FETCH),
    queueOptions: {
      autoDelete: true,
    },
  })
  async fetch(pipeline: Pipeline): Promise<string | null | Nack> {
    const unlock = await this.redisMutexService.lock(
      `repo-project-${pipeline.projectId}`,
    );
    try {
      const git = await this.getGit(pipeline.project, undefined, {
        fetch: false,
      });
      await git.fetch(['origin', pipeline.branch, '--depth=100']);
      return getInstanceName();
    } catch (error) {
      this.logger.error({ error, pipeline }, '[fetch] %s', error?.message);
      return new Nack();
    } finally {
      await unlock();
    }
  }

  @RabbitSubscribe({
    exchange: EXCHANGE_PROJECT_FANOUT,
    routingKey: ROUTE_PROJECT_CHANGE,
    queue: QUEUE_REFRESH_REPO,
    queueOptions: {
      autoDelete: true,
      durable: true,
    },
  })
  async refreshRepo([project]: [Project]) {
    this.logger.info({ project }, '[refreshRepo] start');
    const unlock = await this.redisMutexService.lock(
      `repo-project-${project.id}`,
      {
        timeout: null,
      },
    );
    try {
      const path = join(
        this.configService.get<string>('workspaces.root'),
        encodeURIComponent(project.name),
      );
      await rm(path, { recursive: true });
      this.logger.info({ project }, '[refreshRepo] success');
    } catch (error) {
      this.logger.error(
        { project, error },
        '[refreshRepo] failed. $s',
        error.message,
      );
    } finally {
      await unlock();
    }
  }
}
